//*******************************************************************//
//                                                                   //
//   - More Advanced Programming Topics -                            //
//                                                                   //
//   $Author:       TGraupmann         tgraupmann@yahoo.com          //
//                                                                   //
//   $Program:		3DS Animation PART 3                             //
//                                                                   //
//   $Description:	Shows you how to read the .3DS object hierarchy  //
//                                                                   //
//   $Date:	06/01/03                                                 //
//                                                                   //
//*********************************************************************

// This is a compiler directive that includes libraries (For Visual Studio)
// You can manually include the libraries in the "Project->settings" menu under
// the "Link" tab.  You need these libraries to compile this program.
#pragma comment(lib, "opengl32.lib")
#pragma comment(lib, "glu32.lib")
#pragma comment(lib, "glaux.lib")

#include "main.h"										// This includes our header file
#include "3ds.h"										// Include the 3DS header file.
#include "dirent.h"	// File directory browsing
#include "matrix.h" // Allows basic matrix functions
#include "camera.h" // Allows camera usage
#include "ui.h"     // Gives access to UI Variables

// The following FPS rates are the number of frames per second
// we expect to see each second. This is necessary to control
// the frame rates, otherwise computer speed would have control.
#define RDRFPS 30		// Rendering Frames Per Second
#define MODFPS 45		// Model Frames Per Second
#define CAMFPS 30		// Camera Updates Per Second
#define CAM_SCALE 1.5
#define CAMOFFSET -20

//////////////////////////////UPDATE CAMERA\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////
/////	This function updates the position of the camera
/////
//////////////////////////////UPDATE CAMERA\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*

GLvoid UpdateCamera(GLvoid)
{
	// The code below is a bit tricky. Not only is a FPS calculation going
	// on, but we also calculate what is the expected FPS for the elapsed
	// interval that has just past. If our MODCount has exceeded that count
	// then there is no reason to do another update until more time has past.
	
	long t_elapsed = labs(GetTickCount() - GTUI.camClock);
	
	// Every second, reset the elapsed time to maintain FPS ratio
	if(t_elapsed > 1000) // 1 sec
	{
		GTUI.CameraFPS = GTUI.CAMCount;			
		GTUI.CAMCount -= CAMFPS; // Enforce all updates occur
		GTUI.camClock += t_elapsed;
	}	

	// Here we do a little calculation that takes the elapsed time and multiples
	// with the desired FPS to get how many FPS we should have at the elapsed time.	
	
	// Keep updates at the expected FPS
	if((1000 * GTUI.CAMCount) < (t_elapsed * CAMFPS))
	{
		GTUI.Camera.Update(); // Update the camera data		
		GTUI.CAMCount++; // Increase Count for FPS Calculation		
	}
}


//////////////////////////////UPDATE MODEL\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////
/////	This function 
/////
//////////////////////////////UPDATE MODEL\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*

GLvoid UpdateModel(t3DModel* pModel)
{
	// Here we do something a little different than earlier tutorials.
	// We just create a desired FPS to shoot towards. If the user
	// toggled the delay, we just divide the expected FPS by three.
	short SpeedFPS; // Holds the MSEC speed of frame animation   
	if(!GTUI.FrameDelayToggle)
		SpeedFPS = MODFPS; // switch to desired fps
	else
		SpeedFPS = MODFPS / 3; // reduce the number of frames per second

	// The code below is a bit tricky. Not only is a FPS calculation going
	// on, but we also calculate what is the expected FPS for the elapsed
	// interval that has just past. If our MODCount has exceeded that count
	// then there is no reason to do another update until more time has past.

	long t_elapsed = labs(GetTickCount() - pModel->modClock);

	// Every second, reset the elapsed time to maintain FPS ratio
	if(t_elapsed > 1000) // 1 sec
	{
		pModel->ModelFPS = pModel->MODCount;		
		pModel->MODCount -= SpeedFPS; // Enforce all updates occur
		pModel->modClock += t_elapsed;
	}
   
	// Here we do a little calculation that takes the elapsed time and multiples
	// with the desired FPS to get how many FPS we should have at the elapsed time.
	
	// Keep updates at the expected FPS
	if((1000 * pModel->MODCount) < (t_elapsed * SpeedFPS))
	{		
		// If the numberOfFrames is zero, the mod function would cause
		// a divide by zero error. This remedies that problem.
		if(pModel->numberOfFrames)
		{
			// Increase the frame count but mod it by the total amount of frames
			pModel->currentFrame = (pModel->currentFrame) % (pModel->numberOfFrames) + 1;
		}		
		pModel->MODCount++; // Increase Count for FPS Calculation		
   }
   
	// * IMPORTANT *
	//
	// We do NOT want to animation frame 0 of the animation because it screws
	// up the animation sometimes.  If you have rotations in your model in frame 0
	// it will store that rotation in the animation, which messes it up because
	// when you save the model, the vertices already reflect that rotation.  SO, we just
	// ignore frame 0.  If you need frame 0 for some reason, refer to the Quick Notes
	// for more information on how to make it work (Hint, Reset The Transforms for the object).

	// Make sure we skip frame 0 of the animation
	if(pModel->currentFrame == 0) pModel->currentFrame = 1;	
}


///////////////////////////////// DRAW 3D GRID \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////
/////	This function create a simple grid to get a better view of the animation
/////
///////////////////////////////// DRAW 3D GRID \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*

void Draw3DSAxis()
{
	// Turn the lines YELLOW
	glColor3ub(128, 128, 0);

	// Increase the width of the lines to better show the (X, Y, Z) axis
	glLineWidth(5);

	// Start drawing lines
	glBegin(GL_LINES);

			// Do the X Axis
			glVertex3f(-1000, 0.01f, 0);
			glVertex3f(1000, 0.01f, 0);

			// Do the Y Axis
			glVertex3f(0, 0.01f, 0);
			glVertex3f(0, 1000, 0);

			// Do the Z Axis
			glVertex3f(0, 0.01f, -1000);
			glVertex3f(0, 0.01f, 1000);

	// Stop drawing lines
	glEnd();

	// Reset the line size back to default
	glLineWidth(1.0);
}

void Draw3DSGrid()
{
	// Turn the lines BLUE - GREY	
	glColor3ub(128, 128, 255);	

	// Draw a 10x10 grid along the X and Z axis'
	for(int x = -1000; x <= 1000; x += 10)
	{
		// Start drawing some lines
		glBegin(GL_LINES);

			// Do the horizontal lines (along the X)
			glVertex3i(-1000, 0, x);
			glVertex3i(1000, 0, x);

			// Do the vertical lines (along the Z)
			glVertex3i(x, 0, -1000);
			glVertex3i(x, 0, 1000);

		// Stop drawing lines
		glEnd();
	}
}


///////////////////////////////// RENDER SCENE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////
/////	This function renders the entire scene.
/////
///////////////////////////////// RENDER SCENE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*

void RenderScene() 
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	// Clear The Screen And The Depth Buffer
	glLoadIdentity();									// Reset The matrix	

	GTUI.Camera.Look();		
	
	// Move to current position
	glPushMatrix();
	glTranslatef(GTUI.Camera.Position().x, 0, GTUI.Camera.Position().z);	

	// Make the grid move with the camera, although
	// we want the grid to appear like it's moving.
	int x = (int)fabsf(GTUI.Camera.Position().x);
	int z = (int)fabsf(GTUI.Camera.Position().z);
	glTranslatef(float(x%10), 0, 0 );
	glTranslatef(0, 0, float(z%10) );
	Draw3DSGrid();	
	glPopMatrix();

	// Draw X, Y, Z axis in its fixed position
	Draw3DSAxis();

	// Allow glScale to affect normal vertex arrays
	glEnable(GL_NORMALIZE);
	
	if(GTUI.ModelMode)
	{				
		GTUI.c3DModel[GTUI.SelectedModel].ShowModel();		
	}
	else
	{	
		int count = 0;
		glTranslatef(125, 0, -100);		
		for(int ModelCount=0; ModelCount<GTUI.ModelCount; ModelCount++)
		{
			GTUI.c3DModel[ModelCount].ShowModel();
			if(count++ < 5)
				glTranslatef(-50, 0, 0);
			else
			{
				glTranslatef(250, 0, 50);
				count = 0;
			}
		}		
	}	

	SwapBuffers(GLWin.g_hDC);
	GTUI.SignalDLRecompile = 0;
}

// This is just a helper function, to make our cases look cleaner.
// It passes pModel and pObject by reference and sets them to the
// current GTUI.SelectedModel and GTUI.NodeID indices.
void SelModelObject(t3DModel** pModel, t3DObject** pObject)
{	
	if(GTUI.SelectedModel < GTUI.ModelCount)
	{
		*pModel = &GTUI.c3DModel[GTUI.SelectedModel];
		
		if(GTUI.NodeID < (*pModel)->m_NodeHashTable.size())
			*pObject = &(*pModel)->pObject[(*pModel)->m_NodeHashTable[GTUI.NodeID]];
	}
}


///////////////////////////////// WIN PROC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////
/////	This function handles the window messages.
/////
///////////////////////////////// WIN PROC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*

LRESULT CALLBACK WinProc(HWND hWnd,UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	char strTitle[150] = {0}; // Holds the changing caption name
	LONG    lRet = 0;
	PAINTSTRUCT    ps;	
	
	switch (uMsg)
	{
	case WM_SIZE:										// If the window is resized
	if(!GLWin.g_bFullScreen)							// Do this only if we are NOT in full screen
	{
		SizeOpenGLScreen(LOWORD(lParam),HIWORD(lParam)); // LoWord=Width, HiWord=Height
		GetClientRect(hWnd, &GLWin.g_rRect);			 // Get the window rectangle
	}
	break;
	
	case WM_PAINT:										// If we need to repaint the scene
		BeginPaint(hWnd, &ps);						// Init the paint struct		
		EndPaint(hWnd, &ps);							// EndPaint, Clean up
		break;

	// Below we define our controls for this simple tutorial.
	// The controls are: 

	// Left Mouse Button - Changes the Render mode from normal to wireframe.
	// Right Mouse Button - Turns lighting On/Off
	// F1 -- (DISABLED)
	// F2 -- DIAGNOSTIC MODES
	// F3 -- CYCLE BACK THRU SELECTED MODELS
	// F4 -- CYCLE FORWARD THRU SELECTED MODELS
	// F5 -- CYCLE BACK THRU SELECTED OBJECT
	// F6 -- CYCLE FORWARD THRU SELECTED OBJECT
	// F7 -- CYCLE BACK THRU SELECTED VERTEX
	// F8 -- CYCLE FORWARD THRU SELECTED VERTEX
	// F11 -- CYCLE BACK THRU SELECTED POSITION KEY
	// F12 -- CYCLE FORWARD THRU SELECTED POSITION KEY
	// B -- TOGGLE ENVIRONMENT ATTRIB PUSHING/POPING
	// N -- TOGGLE MATERIALS
	// M -- TOGGLE SINGLE/MULTIPLE MODEL MODE
	// L -- TOGGLE DISPLAY LISTS
	// Z -- TOGGLE BLENDING SOURCE MODE
	// X -- TOGGLE BLENDING DESTINATION MODE
	// 1, 2, 3, 4, 5, 6, 7, 8 -- ANIMATION MODES
	// Escape - Quits	

	case WM_LBUTTONDOWN:								// If the left mouse button was clicked
		
		if(GTUI.ViewMode == GL_TRIANGLES)
		{
			GTUI.ViewMode = GL_LINE_STRIP; // Go to line stips
			SetWindowText(GLWin.g_hWnd, "VIEW MODE [ WIREFRAME ]");
		}
		else
		{
			GTUI.ViewMode = GL_TRIANGLES; // Go to triangles
			SetWindowText(GLWin.g_hWnd, "VIEW MODE [ SOLID ]");
		}
		break;

	case WM_RBUTTONDOWN:								// If the right mouse button was clicked.
		
      GTUI.LightingToggle = !GTUI.LightingToggle; // Turn lighting ON/OFF
      if(GTUI.LightingToggle) // If lighting is Toggled
	  {
         glEnable(GL_LIGHTING); // Enable OpenGL lighting
		 SetWindowText(GLWin.g_hWnd, "Lighting [ ENABLED ]");
	  }
      else
	  {
         glDisable(GL_LIGHTING); // Disable OpenGL lighting
		 SetWindowText(GLWin.g_hWnd, "Lighting [ DISABLED ]");
	  }
      break;

	case WM_CHAR:
		{
			// Precalculation the display information so that these
			// cases are easier to look at.			
			t3DModel* pModel = 0;
			t3DObject* pObject = 0;			
			SelModelObject(&pModel, &pObject);

			TCHAR tchar = (TCHAR)wParam; // Check specific asci keys
			switch(tchar) {			

			case 'r':
				// Recompile all of the display lists
				// You might want to do this sometimes when something about
				// your model has changed.
				//
				// You'll need to recompile the display lists if you want to
				// see the diagnostic changes (like switching to wireframe
				// mode or highlighting the selected node).
				//
				// Recompiling introduces a brief pause
				//
				GTUI.SignalDLRecompile = 1;
				SetWindowText(GLWin.g_hWnd, "Display Lists [ RECOMPILED ]");
				break;

			case 'l':
				// Toggle the use of display lists.
				//
				// You will need to return off display lists if you want to
				// run diagnostics and see your changes on the fly without
				// recompiling.
				//
				GTUI.UseDLToggle = !GTUI.UseDLToggle;
				if(GTUI.UseDLToggle)
					SetWindowText(GLWin.g_hWnd, "Display Lists [ ENABLED ]");
				else
					SetWindowText(GLWin.g_hWnd, "Display Lists [ DISABLED ]");
				break;

			case 'o':
				sprintf(strTitle, "Total {Models: %d, Display Lists: %d, Normals: %d, TexVerts: %d, Verts: %d}",
					GTUI.ModelCount,
					GTUI.TotalDisplayLists,
					GTUI.TotalNormals,
					GTUI.TotalTexVerts,
					GTUI.TotalVerts);
				SetWindowText(GLWin.g_hWnd, strTitle);
				break;

			case 'm':
				GTUI.ModelMode = !GTUI.ModelMode;
				if(GTUI.ModelMode)
					SetWindowText(GLWin.g_hWnd, "Single Model Mode");
				else
					SetWindowText(GLWin.g_hWnd, "Multiple Model Mode");				
				break;

			case 'b':
				GTUI.UseAttrib = !GTUI.UseAttrib;
				if(GTUI.UseAttrib)
					SetWindowText(GLWin.g_hWnd, "Materials (Push Attrib) [ON]");
				else
					SetWindowText(GLWin.g_hWnd, "Materials (Push Attrib) [OFF]");
				break;

			case 'n':
				GTUI.UseMaterials = !GTUI.UseMaterials;
				if(GTUI.UseMaterials)
					SetWindowText(GLWin.g_hWnd, "Materials [ON]");
				else
					SetWindowText(GLWin.g_hWnd, "Materials [OFF]");
				break;

			case 'j':
				if(pModel)
				{
					pModel->MasterScale /= 1.1f;
					sprintf(strTitle, "Master Scale: %f", pModel->MasterScale);
					SetWindowText(GLWin.g_hWnd, strTitle);
				}
				break;

			case 'k':
				if(pModel)
				{
					pModel->MasterScale *= 1.1f;
					sprintf(strTitle, "Master Scale: %f", pModel->MasterScale);
					SetWindowText(GLWin.g_hWnd, strTitle);
				}
				break;			

			case '1':				
				GTUI.g_pivotType = 0;
				sprintf(strTitle, "Main Pivot Type -- supports test models 1-8, character & sandharvestor");
				SetWindowText(GLWin.g_hWnd, strTitle);
				break;

			case '2':				
				GTUI.g_pivotType = 1;
				sprintf(strTitle, "Translated Pivot Type -- supports test model 9");
				SetWindowText(GLWin.g_hWnd, strTitle);
				break;			

			case '3':				
				GTUI.g_pivotType = 3;
				sprintf(strTitle, "Translated Pivot Scenarios -- supports test models 10-13");
				SetWindowText(GLWin.g_hWnd, strTitle);
				break;

			case '4':				
				GTUI.g_pivotType = 2;
				sprintf(strTitle, "Dark Basic Pivot -- works slightly");
				SetWindowText(GLWin.g_hWnd, strTitle);
				break;

			case '5':				
				GTUI.g_pivotType = 4;
				sprintf(strTitle, "MATRIX BASED ANIMATION I");
				SetWindowText(GLWin.g_hWnd, strTitle);
				break;

			case '6':				
				GTUI.g_pivotType = 5;
				sprintf(strTitle, "MATRIX BASED ANIMATION II");
				SetWindowText(GLWin.g_hWnd, strTitle);
				break;

			case '7':				
				GTUI.g_pivotType = 6;
				sprintf(strTitle, "MATRIX BASED ANIMATION III");
				SetWindowText(GLWin.g_hWnd, strTitle);
				break;

			case '8':				
				GTUI.g_pivotType = 7;
				sprintf(strTitle, "NO MODEL ANIMATION");
				SetWindowText(GLWin.g_hWnd, strTitle);
				break;

			case 'p':
				// Cycle to parent of object
				if(GTUI.NodeID != 65535)
				{
					if(pObject)
						if(pObject->m_ParentNode)
							GTUI.NodeID = pObject->m_ParentNode->m_NodeID;
				}
				
				SelModelObject(&pModel, &pObject);
				if(pObject)
					sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: NodeID: %u \"%s\" Group: %u", pObject->m_NodeID, pObject->strName, pObject->m_GroupToggle);
				else if(GTUI.NodeID == 65535)
					sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: NodeID: [NO PARENT]");
				else
					sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: NodeID: [OUT OF RANGE]");				
				break;			

			case 'z': case 'Z':
				GTUI.TransparencySrc = (GTUI.TransparencySrc + 1) % 8;
				sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: Transparency Source Mode: %u (Press Z)", GTUI.TransparencySrc);
				if(!GLWin.g_bFullScreen) // Do this only if we are NOT in full screen
					SetWindowText(GLWin.g_hWnd, strTitle); // Put message in window's title caption
				break;

			case 'x': case 'X':
				GTUI.TransparencyDst = (GTUI.TransparencyDst + 1) % 8;
				sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: Transparency Destination Mode: %u (Press X)", GTUI.TransparencyDst);
				if(!GLWin.g_bFullScreen) // Do this only if we are NOT in full screen
					SetWindowText(GLWin.g_hWnd, strTitle); // Put message in window's title caption
				break;			
			}
		}
		break;
	
	
	case WM_KEYDOWN: // Check if we pressed any of the function keys, arrow keys, or numpad keys
		{
			// Precalculation the display information so that these
			// cases are easier to look at.			
			t3DModel* pModel = 0;
			t3DObject* pObject = 0;			
			SelModelObject(&pModel, &pObject);

			switch(wParam) { // Check if we hit a key			

			case VK_ESCAPE:							// If we hit the escape key
				PostQuitMessage(0);					// Send a QUIT message to the window
				break;			

			case VK_F2:
				// Diagnostic modes
				GTUI.DiagMode = (GTUI.DiagMode + 1) % 4;
				sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: Diagnostic Mode: %u (Press F2)", GTUI.DiagMode);

				// switch to single model model
				GTUI.ModelMode = 1;					

				// switch to line mode
				if(GTUI.DiagMode)
					GTUI.ViewMode = GL_LINE_STRIP;
				else
					GTUI.ViewMode = GL_TRIANGLES;				
				break;

			case VK_F3:
				// Cycle backward through each model
				GTUI.SelectedModel--;
				if(GTUI.SelectedModel < 0)
					GTUI.SelectedModel = 0;

				SelModelObject(&pModel, &pObject);
				if(pModel)
				{
					sprintf(strTitle, "Current Model -- %s", pModel->m_FileName);
					SetWindowText(GLWin.g_hWnd, strTitle);
				}
				break;

			case VK_F4:
				// Cycle forward through each model
				GTUI.SelectedModel++;
				if(GTUI.SelectedModel >= GTUI.ModelCount)
					GTUI.SelectedModel = GTUI.ModelCount - 1;

				SelModelObject(&pModel, &pObject);
				if(pModel)
				{
					sprintf(strTitle, "Current Model -- %s", pModel->m_FileName);
					SetWindowText(GLWin.g_hWnd, strTitle);
				}
				break;

			case VK_F5:
				// Cycle forward through each object
				if(GTUI.NodeID != 65535)
					GTUI.NodeID--;

				SelModelObject(&pModel, &pObject);
				if(pObject)
					sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: NodeID: %u \"%s\" Group: %u Pivot {%.2f, %.2f, %.2f}", GTUI.NodeID, pObject->strName, pObject->m_GroupToggle, pObject->vPivot.x, pObject->vPivot.y, pObject->vPivot.z);
				else if(GTUI.NodeID == 65535)
					sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: NodeID: [NO PARENT]");
				else
				{
					GTUI.NodeID = 0;
					sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: NodeID: [OUT OF RANGE]");
				}				
				break;

			case VK_F6:
				// Cycle forward through each object
				GTUI.NodeID++;
				if(GTUI.NodeID >= pModel->m_NodeHashTable.size())
					GTUI.NodeID--;

				SelModelObject(&pModel, &pObject);
				if(pObject)
					sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: NodeID: %u \"%s\" Group: %u Pivot {%.2f, %.2f, %.2f}", GTUI.NodeID, pObject->strName, pObject->m_GroupToggle, pObject->vPivot.x, pObject->vPivot.y, pObject->vPivot.z);
				else if(GTUI.NodeID == 65535)
					sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: NodeID: [NO PARENT]");
				else
				{
					GTUI.NodeID = 0;
					sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: NodeID: [OUT OF RANGE]");
				}				
				break;

			case VK_F7:
				// Cycle backward through each vertex
				GTUI.VertexID--;
				if(GTUI.VertexID == 65535)
					GTUI.VertexID = 0;
				if(pObject)
				{
					if(GTUI.VertexID < pObject->numOfVerts)
						sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: Vertex[%u] of %u Faces: %u {%.2f, %.2f, %.2f}", GTUI.VertexID, pObject->numOfVerts-1, pObject->numOfFaces-1, pObject->pVerts[GTUI.VertexID].x, pObject->pVerts[GTUI.VertexID].y, pObject->pVerts[GTUI.VertexID].z);
					else
					{
						sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: Vertex[%u] of %u Faces: %u OUT OF RANGE", GTUI.VertexID, pObject->numOfVerts-1, pObject->numOfFaces-1);
						GTUI.VertexID = pObject->numOfVerts - 1;
					}
				}
				else
					sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: Vertex[%u] OUT OF RANGE", GTUI.VertexID);
				break;

			case VK_F8:
				// Cycle forward through each vertex
				if(pObject)
				{
					GTUI.VertexID++;
					if(GTUI.VertexID >= pObject->numOfVerts)
						GTUI.VertexID = pObject->numOfVerts - 1;

					if(GTUI.VertexID < pObject->numOfVerts)
						sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: Vertex[%u] of %u Faces: %u {%.2f, %.2f, %.2f}", GTUI.VertexID, pObject->numOfVerts - 1, pObject->numOfFaces - 1, pObject->pVerts[GTUI.VertexID].x, pObject->pVerts[GTUI.VertexID].y, pObject->pVerts[GTUI.VertexID].z);
					else
					{
						sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: Vertex[%u] of %u Faces: %u OUT OF RANGE", GTUI.VertexID, pObject->numOfVerts - 1, pObject->numOfFaces - 1);
						GTUI.VertexID = pObject->numOfVerts - 1;
					}
				}
				else
					sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: Vertex[%u] OUT OF RANGE", GTUI.VertexID);
				break;			

			case VK_F9:
				GTUI.FrameDelayToggle = !GTUI.FrameDelayToggle;
				if(!GTUI.FrameDelayToggle)
					sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: Frame Speed: NORMAL");
				else
					sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: Frame Speed: SLOW");
				break;			
			
			case VK_F11:
				// cycle backward thru the position keys
				GTUI.TransID--;
				if(GTUI.TransID == 65535)
					GTUI.TransID = 0;
				if(pObject)
				{
					if(GTUI.TransID < pObject->vPosition.size())
						sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: Position ID: [%u] of %u, Key: [#%u] {%.2f, %.2f, %.2f}", GTUI.TransID, pObject->vPosition.size()-1, pObject->vPosFrameNumber[GTUI.TransID], pObject->vPosition[GTUI.TransID].x, pObject->vPosition[GTUI.TransID].y, pObject->vPosition[GTUI.TransID].z);
					else
					{
						sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: Position ID: [%u] of %u OUT OF RANGE", GTUI.TransID, pObject->vPosition.size()-1);
						GTUI.TransID = pObject->vPosition.size() - 1;
					}
				}
				else
					sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: pObject OUT OF RANGE");
				break;

			case VK_F12:
				// cycle forward thru the position keys
				if(pObject)
				{
					GTUI.TransID++;
					if(GTUI.TransID >= pObject->vPosition.size())
						GTUI.TransID = pObject->vPosition.size() - 1;

					if(GTUI.TransID < pObject->vPosition.size())
						sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: Position ID: [%u] of %u, Key: [#%u] {%.2f, %.2f, %.2f}", GTUI.TransID, pObject->vPosition.size()-1, pObject->vPosFrameNumber[GTUI.TransID], pObject->vPosition[GTUI.TransID].x, pObject->vPosition[GTUI.TransID].y, pObject->vPosition[GTUI.TransID].z);
					else
					{
						sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: Position ID: [%u] of %u OUT OF RANGE", GTUI.TransID, pObject->vPosition.size()-1);
						GTUI.TransID = pObject->vPosition.size() - 1;
					}
				}
				else
					sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: pObject OUT OF RANGE");
				break;			

			case VK_SPACE:
				sprintf(strTitle, "www.GameTutorials.com - 3DS Animation Part 3: Display FPS: %.2f, Camera FPS: %u", GTUI.RenderFPS, GTUI.CameraFPS);
				if(pModel)
					sprintf(strTitle, "%s, Model FPS: %u", strTitle, pModel->ModelFPS);
				break;			

			default: // None of the trapped keys were hit, display usable key list
				sprintf(strTitle, "Hot Keys: F2, F3, F4, F5, F6, F7, F8, F9, F12, SPACE, UP, DOWN, LEFT, RIGHT, RXZPASDWBNM");
				break;

			} // Closes switch statement
			
			if(!GLWin.g_bFullScreen) // Do this only if we are NOT in full screen
				SetWindowText(GLWin.g_hWnd, strTitle); // Put message in window's title caption
		}

		break; // END OF CASE

    case WM_CLOSE:										// If the window is being closed
        PostQuitMessage(0);								// Send a QUIT Message to the window
        break; 
     
    default:											// Return by default
        lRet = DefWindowProc (hWnd, uMsg, wParam, lParam); 
        break; 
    }	 
 
    return lRet;										// Return by default
}


///////////////////////////////// MAIN GAME LOOP \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////
/////	This function Handles the main game loop
/////
///////////////////////////////// MAIN GAME LOOP \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*

WPARAM MainLoop()
{
	MSG msg;	

	long t_elapsed; // keep track of elapsed time

	// Static render clock variable
	long rdrClock = GetTickCount();
	
	while(1)											// Do our infinate loop
	{													// Check if there was a message
		if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{
			if(msg.message == WM_QUIT)			// If the message wasnt to quit
				break;
			TranslateMessage(&msg);				// Find out what the message does
			DispatchMessage(&msg);				// Execute the message
		}
		else											// if there wasn't a message
		{
		   // Update Camera more often than display to screen      
			UpdateCamera();			

			// Animate the model by going to the next frame
			for(int ModelCount=0; ModelCount<GTUI.ModelCount; ModelCount++)
			{
				UpdateModel(&GTUI.c3DModel[ModelCount]);
			}

			// Yield a bit to other processes
			Sleep(1);

			//Find how much time has elapsed
			long cam_elapsed = labs(GetTickCount() - GTUI.camClock);
			//Find how much time has elapsed
			long mod_elapsed = labs(GetTickCount() - GTUI.c3DModel[0].modClock);

			// Here we do something a little different than earlier tutorials.
			// We just create a desired FPS to shoot towards. If the user
			// toggled the delay, we just divide the expected FPS by three.
			short SpeedFPS; // Holds the MSEC speed of frame animation   
			if(!GTUI.FrameDelayToggle)
				SpeedFPS = MODFPS; // switch to desired fps
			else
				SpeedFPS = MODFPS / 3; // reduce the number of frames per second

			// Render only when Camera and Model expected FPS counts are reached
			if(
				( (1000 * GTUI.CAMCount) > (cam_elapsed * CAMFPS) ) &&
				( (1000 * GTUI.c3DModel[0].MODCount) > (mod_elapsed * SpeedFPS) )
				)
			{
				// Every second, reset the elapsed time to maintain FPS ratio
				t_elapsed = labs(GetTickCount() - rdrClock);

				if(t_elapsed > 1000) // 1 sec
				{
					//RDRCount is for t_elapsed
					//we want X for 1000
					//RDRCount        X
					//--------   =  ------
					//t_elapsed      1000
					GTUI.RenderFPS = 1000.0f / float(t_elapsed) * float(GTUI.RDRCount);

					//Estimate how many renderings should happen in the elapsed time
					short rdrEst = short((t_elapsed - 1000) * GTUI.RenderFPS); // the time between 1 sec and the next rendering
					GTUI.RDRCount = rdrEst / 1000; // reset to the estimate					
					rdrClock += t_elapsed;					
				}

				// Keep updates at the expected FPS
				if((1000 * GTUI.RDRCount) < (t_elapsed * RDRFPS))
				{				
					RenderScene(); // Update the screen					
					GTUI.RDRCount++; // Increase Count for FPS Calculation					
				}
				else
					Sleep(1);
			}
		}
	}	

	DeInit();										// Clean up and free all allocated memory

	return(msg.wParam);							// Return from the program
}


///////////////////////////////// WIN MAIN \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////
/////	This function handles registering and creating the window.
/////
///////////////////////////////// WIN MAIN \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hprev, PSTR cmdline, int ishow)
{	
	HWND hWnd;

	// Check if we want full screen or not
	if(MessageBox(NULL, "Click Yes to go to full screen (Recommended)", "Options", MB_YESNO | MB_ICONQUESTION) == IDNO)
		GLWin.g_bFullScreen = false;
	else
		GLWin.g_bFullScreen = true;	
	
	// Load our .3DS file into our model structures
	
	GTUI.ModelCount = 0; // Keep a count of the models loaded	
	float scale = 20.0f; // Set a default scale

	///*
	
	// Dark basic test models:
	//

	// Automatically read thru all the 3d models in
	// the dark basic directory. Skip every ten or so.
	DIR *pDIR;
	DIR *cDIR;
	dirent *pFILE;
	char dir_buffer[255];
	char folder_buffer[255];
	char full_path[255];
	int count = 0;

	sprintf(dir_buffer, "/Program Files/Dark Basic Software/DarkMATTER/Models/3DS/People");
	//sprintf(dir_buffer, "/Program Files/Dark Basic Software/DarkMATTER/Models/3DS/Vehicles");	
	//sprintf(dir_buffer, "/Program Files/Dark Basic Software/DarkMATTER/Models/3DS/SkyBoxes");
	pDIR = opendir(dir_buffer);

	while(pFILE = readdir(pDIR))
	{
		//FIND DIRECTORIES IN DARK BASIC FOLDER
		sprintf(folder_buffer, "%s/%s\0", dir_buffer, pFILE->d_name );		
		cDIR = opendir(folder_buffer);
		while(pFILE = readdir(cDIR))
		{
			//FIND MODELS IN NESTED DARK BASIC FOLDER
			int length = strlen(pFILE->d_name);
			if(length > 4) // can only load .3ds extensions
			{				
				if( (toupper(pFILE->d_name[length-4]) == '.') &&
					(toupper(pFILE->d_name[length-3]) == '3') &&
					(toupper(pFILE->d_name[length-2]) == 'D') &&
					(toupper(pFILE->d_name[length-1]) == 'S') )
				{
					if(!(count%10)) // import every nth model
					{
						sprintf(full_path, "%s/%s\0", folder_buffer, pFILE->d_name);
						GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], full_path);
						GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = scale;
						Sleep(10); // yield between loading models						
					}
					count++;
				}				
			}
			Sleep(1); // yield to other processes while browsing dir
		}
	}
	if(pDIR)
		closedir(pDIR);

	//*/
	

	///*	
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/character.3ds");
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 0.5f; // default is stored in the (.3DS) file
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/trans_test01.3ds");
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 0.05f; // default is stored in the (.3DS) file
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/trans_test02.3ds");
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 0.1f; // default is stored in the (.3DS) file
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/trans_test03.3ds");
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 0.05f; // default is stored in the (.3DS) file
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/trans_test04.3ds");
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 0.1f; // default is stored in the (.3DS) file
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/trans_test05.3ds");
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 0.05f; // default is stored in the (.3DS) file
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/trans_test06.3ds");
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 0.1f; // default is stored in the (.3DS) file
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/trans_test07.3ds");
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 0.05f; // default is stored in the (.3DS) file
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/trans_test08.3ds");
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 0.1f; // default is stored in the (.3DS) file
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/trans_test09.3ds");
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 0.05f; // default is stored in the (.3DS) file
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/trans_test10.3ds");
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 0.15f; // default is stored in the (.3DS) file
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/trans_test11.3ds");
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 0.05f; // default is stored in the (.3DS) file
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/trans_test12.3ds");
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 0.05f; // default is stored in the (.3DS) file
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/trans_test13.3ds");
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 0.05f; // default is stored in the (.3DS) file
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/trans_test14.3ds");
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 0.05f; // default is stored in the (.3DS) file
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/trans_test15.3ds");
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 0.05f; // default is stored in the (.3DS) file
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/sandharvestor2.3ds");	
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 4.0f; // default is stored in the (.3DS) file	
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/jammerv3_3.3ds");
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 0.32f; // default is stored in the (.3DS) file
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/fruit1.3ds");
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 0.025f; // default is stored in the (.3DS) file
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/scout2.3ds");
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 0.15f; // default is stored in the (.3DS) file
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/saudor01.3ds");
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 0.04f; // default is stored in the (.3DS) file
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/tree.3ds");
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 10.0f; // default is stored in the (.3DS) file	
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/orbnav01_sm.3ds");
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 6.0f; // default is stored in the (.3DS) file	
	GTUI.cLoad3ds.Import3DS(&GTUI.c3DModel[GTUI.ModelCount++], "data/face.3DS");
	GTUI.c3DModel[GTUI.ModelCount - 1].MasterScale = 2.5f; // default is stored in the (.3DS) file	
	//*/
	
	// Create our window with our function we create that passes in the:
	// Name, width, height, any flags for the window, if we want fullscreen of not, and the hInstance
	hWnd = CreateMyWindow("www.GameTutorials.com - 3DS Animation Part 3", SCREEN_WIDTH, SCREEN_HEIGHT, 0, GLWin.g_bFullScreen, hInstance);

	// If we never got a valid window handle, quit the program
	if(hWnd == NULL) return true;

	// INIT OpenGL
	Init(hWnd);													

	// Run our message loop and after it's done, return the result
	return MainLoop();						
}


/////////////////////////////////////////////////////////////////////////////////
//
// * QUICK NOTES * 
//
// This tutorial is Part III in the 3DS Animation series. This release
// is a minor update to the last 3DS Animation tutorial.
//
// Coverage:
// Part I -- Covered loading 3DS Animations for objects lacking parents.
//
// Part II -- Covered loading 3DS Animations for objects with object parents.
//
// Part III -- This tutorial. Covers loading 3DS Animations for objects that
//             have group parents.
//
// This tutorial remains simple yet again, by avoiding the use of Matrices.
//
//
// All window handling code was moved to init.cpp. Diagnostic code
// was moved to diagnostics.cpp.
//
// Keyboard handling and mouse handling implementation remains in main.cpp
// within WinProc.
//
// Global variables were moved to main.h and assigned into structures
// to keep the data more organized.
//
// The existing global variables were moved into GTUI_Interface and
// GT3DS_Handler. GTUI_Interface holds all the variables that can be
// manipulated through the UI. GT3DS_Handler holds all the variables
// respective to the Filename of the model, the actual model, textures
// for the model and the CLoad3DS structure.
// 
// Window handling variables are now
// stored in the GLWindow structure in main.h.
//
// The original CVector3, CVector2, tMaterialInfo, t3DObject, and
// t3DModel are all still included in main.h.
//
// FPS code was added so that rendering does not exceed 30 FPS.
//
// The implementation for moving the camera, moving lights, advancing
// frames was moved from the rendering code, so that it would no
// longer be dependent on rendering. Additionally, FPS calculations
// were added to the update functions so that things would happen
// at a constant rate.
// 
// Enjoy!
//
//
// Tim Graupmann
// Game Programmer
// tgraupmann@yahoo.com
// http://tgraupmann.hypermart.net
//
//