#include <windows.h>
#include <gl\gl.h>			// Header File For The OpenGL32 Library

#include "models.h"
#include "diagnostics.h"
#include "ui.h"
#include "animation.h"
#include "main.h"


// This holds our model information.  This should also turn into a robust class.
// We use STL's (Standard Template Library) vector class to ease our link list burdens. :)

// This function shows the given model using display lists
// or not, depending on the UI variables that are set.	
void t3DModel::ShowModel(void)
{	
	glPushMatrix();		
	glScalef(MasterScale, MasterScale, MasterScale);
	
	// Since we know how many objects our model has, go through each of them.
	for(int i = 0; i < numOfObjects; i++)
	{
		// Make sure we have valid objects just in case. (size() is in the vector class)
		if(pObject.size() <= 0) break;

		// Get the current object that we are displaying
		t3DObject *pObj = &pObject[i];		

		// Commented out so that the group pivot can be drawn
		//if(pObj->m_GroupToggle)
			//continue;		
		
		// Push on a new matrix for every object so we don't translate/rotate/scale for another
		glPushMatrix();
		
		// Apply all the pivot diagnostics (temporary)
		PivotDiagnostics(this, pObj, currentFrame);

		// Animate the current node
		switch(GTUI.g_pivotType)
		{
		case 0:
			AnimateModel_Method_1(this, pObj);
			break;
		case 1:
			AnimateModel_Method_2(this, pObj);
			break;
		case 2:
			AnimateModel_Method_3(this, pObj);
			break;
		case 3:
			AnimateModel_Method_4(this, pObj);
			break;
		case 4:
			AnimateModel_Method_5(this, pObj);
			break;
		case 5:
			{
			int Tier = 0;
			AnimateModel_Method_6(this, pObj, &Tier);
			}
			break;
		case 6:
			{
			int Tier = 0;
			AnimateModel_Method_7(this, pObj, &Tier);
			}
			break;
		}

		PivotDiagnostics(this, pObj, currentFrame);

		// Apply the animation tracks of group models, but
		// Skip drawing the verteces of group keys
		if(pObj->m_GroupToggle || !pObj->pVerts)
		{
			// This allows the pivot point of the group to be drawn.
			// This truely belongs above. (Where it's commented out).
			glPopMatrix();			
			continue;
		}

		// Toggle pushing/poping attributes each material change
		if(GTUI.UseAttrib)
		{
			glPushAttrib(GL_ALL_ATTRIB_BITS);
		}		

		if(GTUI.UseDLToggle) // If we want to use display lists
		{
			// Render the Display List
			glCallList(pObj->m_DisplayListID);
		}
		else
		{
			// Toggle using materials for each object
			if(GTUI.UseMaterials)
			{
				// Apply the material colors to the object
				SetMaterial(pObj);
			}
			// Just draw the object without the display list
			// optimizations to compare against.			
			DrawObject(pObj);
		}

		// Toggle pushing/poping attributes each material change
		if(GTUI.UseAttrib)
		{
			glPopAttrib();
		}

		// If we've signaled we want the display list to recompile,
		// then recompile the display lists. This does not get
		// displayed.
		//
		// The only reason that this call is in the rendering
		// function is because it is conveniently already
		// looping thru each object. This would be better put
		// in a function that can be called from the mainloop.
		//
		if(GTUI.SignalDLRecompile)
		{
			// In case the display list needs to be recompiled
			if(pObj->m_DisplayListID)
				CreateDisplayList(pObj, pObj->m_DisplayListID);
		}

		// Pop the current matrix and start fresh for another object
		glPopMatrix();		
	}

    glPopMatrix();    
}

// This function sets the material/textures for the object.
void t3DModel::SetMaterial(t3DObject *pObject)
{	
	GLfloat TRANSPARENCY;
	TRANSPARENCY = pMaterials[pObject->materialID].m_pTRANSPARENCY;

	// Set blending function
	if(TRANSPARENCY)
	{
		glBlendFunc(GTUI.TransparencyTable[GTUI.TransparencySrc],
			GTUI.TransparencyTable[GTUI.TransparencyDst]);
		glEnable(GL_BLEND);
	}
	else
	{
		glBlendFunc(GL_ONE, GL_ZERO);
		glEnable(GL_BLEND);
	}
	
	GLfloat AMBIENT[4];		
	memcpy(AMBIENT, pMaterials[pObject->materialID].m_pAMBIENT, sizeof(float)*4);
	AMBIENT[3] = TRANSPARENCY;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, AMBIENT);
	
	GLfloat DIFFUSE[4];
	memcpy(DIFFUSE, pMaterials[pObject->materialID].m_pDIFFUSE, sizeof(float)*4);
	DIFFUSE[3] = TRANSPARENCY;
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, DIFFUSE);
	
	GLfloat SPECULAR[4];
	memcpy(SPECULAR, pMaterials[pObject->materialID].m_pSPECULAR, sizeof(float)*4);
	SPECULAR[3] = TRANSPARENCY;
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, SPECULAR);
	
	GLfloat SHININESS;
	SHININESS = pMaterials[pObject->materialID].m_pSHININESS;
	glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, &SHININESS);		

	// Check to see if this object has a texture map, if so bind the texture to it.
	if(pObject->bHasTexture) {
		// Turn on texture mapping and turn off color
		glEnable(GL_TEXTURE_2D);
		glColor4f(1, 1, 1, TRANSPARENCY); // default color
		// Bind the texture map to the object by it's materialID
		glBindTexture(GL_TEXTURE_2D, GTUI.Texture[pMaterials[pObject->materialID].texureId]);
	}
	else
	{
		// Turn off texture mapping and turn on color
		glDisable(GL_TEXTURE_2D);			

		if(pMaterials.size() && pObject->materialID >= 0)
		{
			// Get and set the color that the object is, since it must not have a texture
			float *pColor = pMaterials[pObject->materialID].m_pDIFFUSE;
			
			// Assign the current color to this model
			glColor4f(pColor[0], pColor[1], pColor[2], TRANSPARENCY);
		}
		else
		{
			// Reset the color to normal instead
			glColor4ub(255, 255, 255, 255);
		}
	}	

	// When pivot diagnostics are on, this allows the
	// current node to appear highlighted.
	if(GTUI.DiagMode % 3)
	{
		float *pColor = pMaterials[pObject->materialID].m_pDIFFUSE;

		if(IsParentQuery(pObject, GTUI.NodeID))
			glColor4f(
				1.0f-pColor[0],
				1.0f-pColor[1]/2.0f,
				pColor[2]/2.0f,
				TRANSPARENCY);			
	}
}

// This function draws the object and uses all the drawing
// calls whether in a display list or not.
void t3DModel::DrawObject(t3DObject *pObject)
{
	// -------[[[[Draw the mesh]]]]------ //
	// **** New vertex array method ***** //

	// Make sure we have texture coordinates to render
	if(pObject->pTexVerts) 
	{
		// Turn on the texture coordinate state
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);

		// Point OpenGL to our texture coordinate array.
		// We have them in a pair of 2, of type float and 0 bytes of stride between them.
		glTexCoordPointer(2, GL_FLOAT, 0, pObject->pTexVerts);

		GTUI.TotalTexVerts++;
	}

	// Make sure we have vertices to render
	if(pObject->pVerts)
	{
		// Turn on the vertex array state
		glEnableClientState(GL_VERTEX_ARRAY);

		// Point OpenGL to our vertex array.  We have our vertices stored in
		// 3 floats, with 0 stride between them in bytes.
		glVertexPointer(3, GL_FLOAT, 0, pObject->pVerts);

		GTUI.TotalVerts++;
	}

	// Make sure we have normals to render
	if(pObject->pNormals)
	{
		// Turn on the normals state
		glEnableClientState(GL_NORMAL_ARRAY);

		// Point OpenGL to our normals array.  We have our normals
		// stored as floats, with a stride of 0 between.
		glNormalPointer(GL_FLOAT, 0, pObject->pNormals);

		GTUI.TotalNormals++;
	}

	// Here we pass in the indices that need to be rendered.  We want to
	// render them in triangles, with numOfFaces * 3 for indice count,
	// and the indices are of type UINT (important).
	if(pObject->pIndices)
	{
		glDrawElements(GTUI.ViewMode,   pObject->numOfFaces * 3, 
				        GL_UNSIGNED_INT, pObject->pIndices);		
	}	
}

// This function creates a display list using displayListOffset as
// the index to be used. The function then calls DrawObject and
// compiles all the drawing commands therein.
void t3DModel::CreateDisplayList(t3DObject *pObject, int displayListOffset)
{
	// Put the display list ID into the model object
	pObject->m_DisplayListID = displayListOffset;

	// Signal the beginning of the Display List
	glNewList(pObject->m_DisplayListID, GL_COMPILE);

	if(GTUI.UseMaterials)
	{
		// Apply the material colors to the object
		SetMaterial(pObject);
	}	

	// Draw the object and compile into list
	DrawObject(pObject);

	// Signal the End of the Display List
	glEndList();

}

// Constructor
t3DModel::t3DModel()
{
	numOfObjects = 0;
	numOfMaterials = 0;
	numberOfFrames = 0;
	currentFrame = 0;
	pMaterials.clear();
	pObject.clear();
	MasterScale = 0;
	m_NodeHashTable.clear();
	modClock = GetTickCount();
	MODCount = 0;
	ModelFPS = 0;
	memset(m_FileName, 0, sizeof(char) * 255);
}

// Copy constructor
t3DModel::t3DModel(const t3DModel& copy)
{
	numOfObjects = copy.numOfObjects;
	numOfMaterials = copy.numOfMaterials;
	numberOfFrames = copy.numberOfFrames;
	currentFrame = copy.currentFrame;
	// Copy vector list -- pMaterials
	for(unsigned short i = 0; i < copy.pMaterials.size(); i++)
		pMaterials.push_back( copy.pMaterials[i] );
	// Copy vector list -- pObject
	for(i = 0; i < copy.pObject.size(); i++)
		pObject.push_back( copy.pObject[i] );		
	MasterScale = copy.MasterScale;
	// Copy vector list -- m_NodeHashTable
	for(i = 0; i < copy.m_NodeHashTable.size(); i++)
		m_NodeHashTable.push_back( copy.m_NodeHashTable[i] );
	modClock = copy.modClock;
	MODCount = copy.MODCount;
	ModelFPS = copy.ModelFPS;
	memcpy(m_FileName, copy.m_FileName, sizeof(char) * 255);
}

// Destructor
t3DModel::~t3DModel()
{
	numOfObjects = 0;
	numOfMaterials = 0;
	numberOfFrames = 0;
	currentFrame = 0;
	pMaterials.clear();
	pObject.clear();
	MasterScale = 0;
	m_NodeHashTable.clear();
	modClock = GetTickCount();
	MODCount = 0;
	ModelFPS = 0;
	memset(m_FileName, 0, sizeof(char) * 255);
}